import sys
from collections import defaultdict
from dataclasses import dataclass
from typing import *

import cbor2
import yaml
from zuper_ipce.json2cbor import read_cbor_or_json_objects

from aido_schemas import (
    RobotState,
    RobotObservations,
    Duckiebot1Observations,
    SetRobotCommands,
)
from duckietown_world import SE2Transform, SampledSequence, DuckietownMap, draw_static
from duckietown_world.rules import evaluate_rules
from duckietown_world.rules.rule import make_timeseries, RuleEvaluationResult
from duckietown_world.seqs.tsequence import SampledSequenceBuilder
from duckietown_world.svg_drawing.draw_log import (
    SimulatorLog,
    timeseries_actions,
    RobotTrajectories,
)
from duckietown_world.svg_drawing.misc import TimeseriesPlot
from duckietown_world.world_duckietown.types import SE2v
from duckietown_world.world_duckietown.utils import get_velocities_from_sequence

from zuper_ipce.conv_object_from_ipce import object_from_ipce
from . import logger


@dataclass
class LogData:
    objects: List[dict]


def log_summary(filename) -> LogData:
    objects = []
    with open(filename, "rb") as f:
        counts = defaultdict(lambda: 0)
        sizes = defaultdict(lambda: 0)
        for ob in read_cbor_or_json_objects(f):
            topic = ob["topic"]
            size_ob = len(cbor2.dumps(ob))
            counts[topic] += 1
            sizes[topic] += size_ob
            objects.append(ob)

    ordered = sorted(counts, key=lambda x: sizes[x], reverse=True)
    for topic in ordered:
        count = counts[topic]
        size_mb = sizes[topic] / (1024 * 1024.0)
        logger.info("topic %25s: %4s messages  %.2f MB" % (topic, count, size_mb))
    return LogData(objects)


def read_topic2(ld: LogData, topic):
    for ob in ld.objects:
        if ob["topic"] == topic:
            yield ob


def read_map_info(ld: LogData) -> DuckietownMap:
    m = list(read_topic2(ld, "set_map"))
    if not m:
        msg = "Could not find set_map"
        raise Exception(msg)
    m = m[0]
    map_data_yaml = m["data"]["map_data"]
    map_data = yaml.load(map_data_yaml, Loader=yaml.SafeLoader)
    duckietown_map = construct_map(map_data)
    return duckietown_map


import numpy as np


def read_perfomance(ld: LogData) -> Dict[str, RuleEvaluationResult]:
    sb = SampledSequenceBuilder[float]()
    sb.add(0, 0)
    sequences: Dict[str, SampledSequenceBuilder] = defaultdict(
        lambda: SampledSequenceBuilder[float]()
    )

    for i, ob in enumerate(read_topic2(ld, "timing_information")):
        # ob = ipce_to_object(ob['data'], {}, {})
        phases = ob["data"]["phases"]
        phases.pop("$schema")
        for p, f in phases.items():
            sequences[p].add(t=i, v=f)

    evr = RuleEvaluationResult(None)
    for p, sb in sequences.items():
        seq = sb.as_sequence()
        total = float(np.mean(seq.values))
        evr.set_metric((p,), total=total, incremental=seq, cumulative=None)

    return {"performance": evr}


def read_trajectories(ld: LogData) -> Dict[str, RobotTrajectories]:
    rs = list(read_topic2(ld, "robot_state"))
    if not rs:
        msg = "Could not find robot_state"
        raise Exception(msg)
    robot_names = set([r["data"]["robot_name"] for r in rs])
    logger.info(f"robot_names: {robot_names}")

    robot2trajs = {}
    for robot_name in robot_names:
        ssb_pose = SampledSequenceBuilder[SE2Transform]()
        ssb_pose_SE2 = SampledSequenceBuilder[SE2v]()
        ssb_actions = SampledSequenceBuilder[Any]()
        ssb_wheels_velocities = SampledSequenceBuilder[Any]()
        # ssb_velocities = SampledSequenceBuilder[Any]()
        for r in rs:

            robot_state = cast(RobotState, object_from_ipce(r["data"]))
            if robot_state.robot_name != robot_name:
                continue

            pose = robot_state.state.pose
            # velocity = robot_state.state.velocity
            last_action = robot_state.state.last_action
            wheels_velocities = robot_state.state.wheels_velocities

            t = robot_state.t_effective
            ssb_pose_SE2.add(t, pose)
            ssb_pose.add(t, SE2Transform.from_SE2(pose))
            ssb_actions.add(t, last_action)
            ssb_wheels_velocities.add(t, wheels_velocities)
            # ssb_velocities.add(t, velocity)

        seq_velocities = get_velocities_from_sequence(ssb_pose_SE2)
        observations = read_observations(ld, robot_name)
        commands = read_commands(ld, robot_name)

        robot2trajs[robot_name] = RobotTrajectories(
            ssb_pose.as_sequence(),
            ssb_actions.as_sequence(),
            ssb_wheels_velocities.as_sequence(),
            seq_velocities,
            observations=observations,
            commands=commands,
        )
    return robot2trajs


from duckietown_world.world_duckietown import DB18, construct_map


def read_observations(ld: LogData, robot_name):
    ssb = SampledSequenceBuilder[bytes]()
    obs = list(read_topic2(ld, "robot_observations"))
    last_t = None
    for ob in obs:
        ro = cast(RobotObservations, object_from_ipce(ob["data"]))
        if ro.robot_name != robot_name:
            continue
        do: Duckiebot1Observations = ro.observations

        t = ro.t_effective
        camera = do.camera.jpg_data

        if last_t != t:
            ssb.add(t, camera)
        last_t = t
    res = ssb.as_sequence()
    return res


def read_commands(ld: LogData, robot_name):
    ssb = SampledSequenceBuilder[SetRobotCommands]()
    obs = list(read_topic2(ld, "set_robot_commands"))
    last_t = None
    for ob in obs:
        ro = cast(SetRobotCommands, object_from_ipce(ob["data"]))
        if ro.robot_name != robot_name:
            continue
        t = ro.t_effective
        if last_t != t:
            ssb.add(ro.t_effective, ro.commands)
        last_t = t
    seq = ssb.as_sequence()
    if len(seq) == 0:
        msg = f'Could not find any robot_commands in the log for robot "{robot_name}".'
        logger.warning(msg)
    return seq


def read_simulator_log_cbor(ld: LogData) -> SimulatorLog:
    render_time = read_perfomance(ld)
    duckietown_map = read_map_info(ld)
    robots = read_trajectories(ld)

    for robot_name, trajs in robots.items():
        robot = DB18()
        duckietown_map.set_object(robot_name, robot, ground_truth=trajs.pose)

    return SimulatorLog(
        duckietown=duckietown_map, robots=robots, render_time=render_time
    )


def read_and_draw(fn: str, output: str):
    ld = log_summary(fn)

    logger.info("Reading logs...")
    log0 = read_simulator_log_cbor(ld)
    logger.info("...done")

    robot_main = "ego"
    if not robot_main in log0.robots:
        msg = f"Cannot find robot {robot_main}"
        raise Exception(msg)
    log = log0.robots[robot_main]

    if log.observations:
        images = {"observations": log.observations}
    else:
        images = None
    duckietown_env = log0.duckietown
    timeseries = {}

    logger.info("Computing timeseries_actions...")
    timeseries.update(timeseries_actions(log))
    logger.info("Computing timeseries_wheels_velocities...")
    timeseries.update(timeseries_wheels_velocities(log.commands))
    logger.info("Computing timeseries_robot_velocity...")
    timeseries.update(timeseries_robot_velocity(log.velocity))
    logger.info("Evaluating rules...")
    interval = SampledSequence.from_iterator(enumerate(log.pose.timestamps))
    evaluated = evaluate_rules(
        poses_sequence=log.pose,
        interval=interval,
        world=duckietown_env,
        ego_name=robot_main,
    )
    evaluated.update(log0.render_time)

    for k, v in evaluated.items():
        for kk, vv in v.metrics.items():
            logger.info("%20s %20s %s" % (k, kk, vv))
    timeseries.update(make_timeseries(evaluated))
    logger.info("Drawing...")
    draw_static(duckietown_env, output, images=images, timeseries=timeseries)
    logger.info("...done.")
    return evaluated


def timeseries_wheels_velocities(log_commands):
    timeseries = {}
    sequences = {}
    sequences["motor_left"] = log_commands.transform_values(
        lambda _: _.wheels.motor_left, float
    )
    sequences["motor_right"] = log_commands.transform_values(
        lambda _: _.wheels.motor_right, float
    )
    timeseries["pwm_commands"] = TimeseriesPlot(
        "PWM commands", "pwm_commands", sequences
    )
    return timeseries


import geometry


def timeseries_robot_velocity(log_velocity):
    timeseries = {}
    sequences = {}

    logger.info(log_velocity)

    def speed(x) -> float:
        l, omega = geometry.linear_angular_from_se2(x)
        return l[0]

    def omega(x) -> float :
        l, omega = geometry.linear_angular_from_se2(x)
        return omega

    sequences["linear_speed"] = log_velocity.transform_values(lambda _: speed(_), float)
    sequences["angular_velocity"] = log_velocity.transform_values(lambda _: omega(_), float)
    logger.info("linear speed: %s" % sequences["linear_speed"])
    logger.info("angular velocity: %s" % sequences["angular_velocity"])
    timeseries["velocity"] = TimeseriesPlot("Velocities", "velocities", sequences)
    return timeseries


def aido_log_draw_main():
    read_and_draw(sys.argv[1], "test")


if __name__ == "__main__":
    read_and_draw(sys.argv[1], "test")



from . import InteractionProtocol
from .protocol_agent import EpisodeStart
from .schemas import JPGImage

__all__ = ["protocol_image_filter", "protocol_image_source"]

protocol_image_filter = InteractionProtocol(
    description="""An image filter. Takes an image, returns an image.""",
    inputs={"image": JPGImage, "episode_start": EpisodeStart},
    outputs={"image": JPGImage, "episode_start": EpisodeStart},
    language="""
        (in:episode_start ; out:episode_start ; (in:image ; out:image)*)*
        """,
)

protocol_image_source = InteractionProtocol(
    description="""

An abstraction over logs of images. 

It emits a series of EpisodeStart followed by a set of images.

    """,
    inputs={"next_image": type(None), "next_episode": type(None)},
    outputs={
        "image": JPGImage,
        "episode_start": EpisodeStart,
        "no_more_images": type(None),
        "no_more_episodes": type(None),
    },
    language="""
                (
                    in:next_episode ; (
                        out:no_more_episodes | 
                        (out:episode_start ;
                            (in:next_image ; (out:image | out:no_more_images))*)
                    )
                )*            
            """,
)



from aido_schemas import protocol_image_source
from comptests import comptest
from zuper_nodes import OutputProduced, Enough, Unexpected
from zuper_nodes_tests.test_protocol import assert_seq


@comptest
def test_proto_image_source():
    l0 = protocol_image_source.language
    seq = [OutputProduced("next_image")]
    assert_seq(l0, seq, (Unexpected,), Unexpected)


@comptest
def test_run_docker0():
    pass


@comptest
def test_run_docker1():
    pass


@comptest
def test_run_docker2():
    pass


@comptest
def test_run_docker3():
    pass


@comptest
def test_run_docker4():
    pass


@comptest
def test_run_docker5():
    pass


@comptest
def test_run_docker6():
    pass


@comptest
def test_run_docker7():
    pass



from dataclasses import dataclass
from typing import Any

from . import InteractionProtocol

__all__ = ["EpisodeStart", "protocol_agent"]


@dataclass
class EpisodeStart:
    """ Marker for the start of an episode. """

    episode_name: str


protocol_agent = InteractionProtocol(
    description="""

Generic protocol for an agent that receives "observations" and responds 
with "commands".

"episode_start" marks the beginning of an episode.  

    """.strip(),
    inputs={
        "observations": Any,
        "seed": int,
        "get_commands": type(None),
        "episode_start": EpisodeStart,
    },
    outputs={"commands": Any},
    language="""
            in:seed? ;
            (   in:episode_start ; 
                (in:observations | 
                    (in:get_commands ; out:commands)
                 )* 
            )*
        """,
)



import sys

from aido_schemas.utils_drawing import read_simulator_log_cbor, log_summary
from duckietown_world.svg_drawing.draw_log import SimulatorLog
from procgraph import Generator, Block, register_model_spec, pg


class CBORRead(Generator):
    """
    """

    Block.alias("cborread")
    Block.output("image")
    Block.config("filename", "CBOR file to read")
    Block.config("robot_name", "robot name", default="ego")

    # noinspection PyAttributeOutsideInit
    def init(self):
        fn = self.get_config("filename")
        self.ld = log_summary(fn)
        self.log: SimulatorLog = read_simulator_log_cbor(self.ld)
        self.i = 0
        log = self.log.robots[self.config.robot_name]
        self.n = len(log.observations)

    def next_data_status(self):
        log = self.log.robots[self.config.robot_name]
        if self.i < self.n:
            return True, log.observations.timestamps[self.i]
        else:
            return False, None

    def update(self):
        i = self.i
        log = self.log.robots[self.config.robot_name]
        timestamp = log.observations.timestamps[i]
        value = log.observations.values[i]
        self.set_output("image", value=value, timestamp=timestamp)

        self.i += 1

    def finish(self):
        pass


def make_video1(fn, output):
    register_model_spec(
        """
    --- model video_aido
    config output
    config filename
    

    |cborread filename=$filename| --> |jpg2rgb| -> rgb
    rgb -> |identity| -> retimed
    # rgb --> |rewrite_timestamps interval=$factor| -> retimed
    retimed --> |info|
    retimed -> |mencoder quiet=1 file=$output timestamps=0|

        """
    )

    pg("video_aido", dict(filename=fn, output=output))


def aido_log_video_main():
    make_video1(sys.argv[1], "out-aido-log-video.mp4")


if __name__ == "__main__":
    make_video1(sys.argv[1], "out-aido-log-video.mp4")



from setuptools import setup


def get_version(filename: str):
    import ast
    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError('No version found in %r.' % filename)
    if version is None:
        raise ValueError(filename)
    return version


version = get_version(filename='src/aido_schemas/__init__.py')

line = 'daffy'

setup(
    name=f'aido-protocols-{line}',
    version=version,
    keywords='',
    package_dir={'': 'src'},
    packages=[
        'aido_schemas',
        'aido_schemas_tests',
    ],
    install_requires=[
        'compmake',
        'pyparsing',
        'PyContracts',
        'networkx<=2.2',
        'termcolor',
        'cbor2',
        'pydot',
        'zuper-ipce-z5>=5.1.0',
        'zuper-nodes-z5>=5.0.7',
    ],
    entry_points={
        'console_scripts': [
            'aido-log-draw=aido_schemas.utils_drawing:aido_log_draw_main',
            'aido-log-video=aido_schemas.utils_video:aido_log_video_main',
        ],
    },
)



__version__ = "5.0.18"


import logging

logging.basicConfig()
logger = logging.getLogger("aido-protocols")
logger.setLevel(logging.DEBUG)
logger.info(f"aido-protocols {__version__}")

from zuper_nodes import InteractionProtocol, particularize

from zuper_nodes_wrapper import wrap_direct, Context

_ = wrap_direct
_ = Context
_ = InteractionProtocol
_ = particularize

from .protocol_agent import *
from .protocol_simulator import *
from .protocols import *
from .schemas import *



from dataclasses import dataclass

from . import particularize
from .protocol_agent import protocol_agent
from .protocol_simulator import (
    RobotName,
    protocol_simulator,
    SetRobotCommands,
    RobotObservations,
)

__all__ = [
    "PWMCommands",
    "JPGImage",
    "Duckiebot1Observations",
    "Duckiebot1Commands",
    "LEDSCommands",
    "RGB",
    "DB18SetRobotCommands",
    "DB18RobotObservations",
    "protocol_agent_duckiebot1",
    "protocol_simulator_duckiebot1",
]


@dataclass
class PWMCommands:
    """
        PWM commands are floats between -1 and 1.
    """

    motor_left: float
    motor_right: float

    def __post_init__(self):
        self.motor_left = float(self.motor_left)
        self.motor_right = float(self.motor_right)
        m = max(abs(self.motor_left), abs(self.motor_right))
        if m > 1:
            msg = f'Expected values to be between -1 and 1. Obtained {self.motor_left}, {self.motor_right}'
            raise ValueError(msg)



# @dataclass
# class WheelsCmd:
#     """ Kinematic wheels commands. Radiants per second. """
#     vel_left: float
#     vel_right: float


@dataclass
class JPGImage:
    """
        An image in JPG format.

        jpg_data: Bytes of a JPG file
    """

    jpg_data: bytes


@dataclass
class Duckiebot1Observations:
    camera: JPGImage


@dataclass
class RGB:
    r: float
    g: float
    b: float


@dataclass
class LEDSCommands:
    center: RGB
    front_left: RGB
    front_right: RGB
    back_left: RGB
    back_right: RGB


@dataclass
class Duckiebot1Commands:
    wheels: PWMCommands
    LEDS: LEDSCommands


description = """Particularization for Duckiebot1 observations and commands."""
protocol_agent_duckiebot1 = particularize(
    protocol_agent,
    description=description,
    inputs={"observations": Duckiebot1Observations},
    outputs={"commands": Duckiebot1Commands},
)


@dataclass
class DB18SetRobotCommands(SetRobotCommands):
    robot_name: RobotName
    t_effective: float
    commands: Duckiebot1Commands


@dataclass
class DB18RobotObservations(RobotObservations):
    robot_name: RobotName
    t_effective: float
    observations: Duckiebot1Observations


description = """Particularization for Duckiebot1 observations and commands."""
protocol_simulator_duckiebot1 = particularize(
    protocol_simulator,
    description=description,
    inputs={"set_robot_commands": DB18SetRobotCommands},
    outputs={"robot_observations": DB18RobotObservations},
)



from . import InteractionProtocol

__all__ = ["protocol_simple_predictor"]

protocol_simple_predictor = InteractionProtocol(
    description="""

An estimator receives a stream of values and must predict the next value.

    """.strip(),
    inputs={"observations": float, "seed": int, "get_prediction": type(None)},
    outputs={"prediction": float},
    language="""
            in:seed? ;
        
            (in:observations | 
                (in:get_prediction ; out:prediction)
             )*
        """,
)
